{% extends "base.txt.j2" %}
{% block content %}
Tag Name: {{ data.tag|trim }}

Module:

{{ data.module|pprint }}

Traceback:

{{ data.traceback }}

{% if data.build_url %}
Check for more details: {{ data.build_url|trim }}
{% endif %}

{% endblock %}
